/*
 * utmp.c	Write system and radius utmp files.
 *
 * Version:	#(#)utmp.c  1.00  12-Nov-1997  miquels@cistron.nl
 *
 */
#include <stdio.h>
#include <sys/types.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <utmp.h>
#include "server.h"

int update_utmp(int port, struct auth *ai, int islogin)
{
	struct utmp tmp, *utmptr;
	FILE *fp;
	pid_t pid;
	char username[sizeof(ai->login)];
	char *p;
#ifdef BIGUTMP	
	char *m;
	char ip[256];
#endif

	pid = getpid();

	if (lineconf[port].sysutmp && islogin) {
		/*
		 *	See if we need to strip the username.
		 */
		strcpy(username, ai->login);
		if ((mainconf.stripnames) && (islogin > P_INCOM)){
			if (strchr("PCS!L", ai->login[0])) {
				strcpy(username, ai->login + 1);
			} else {
				p = strrchr(username, '.');
				if (p &&
				    (strcmp(p, ".slip")  == 0 ||
				     strcmp(p, ".cslip") == 0 ||
				     strcmp(p, ".ppp")   == 0))
					*p = 0;
			}
		}
		/*
		 *	Init should have setup this for us, so we
		 *	only update the entry if it exists (and only
		 *	at login - init does logout for us).
		 */
		setutent();
		while((utmptr = getutent()) != NULL)
			if (utmptr->ut_pid == pid)
				break;
		if (utmptr) {
			tmp = *utmptr;
			tmp.ut_type = USER_PROCESS;
			tmp.ut_time = time(NULL);
			strncpy(tmp.ut_user, username, UT_NAMESIZE);
			strncpy(tmp.ut_line, lineconf[port].tty,
				UT_LINESIZE);

#ifndef BIGUTMP
			if ((ai->address) && (islogin > P_INCOM)) {
				/*
				 *	Fill in ut_host as
				 *	port:X.55.2
				 */
				char *p = dotted(ai->address);
				int i = 0;

				while(*p && i < 2)
					if (*p++ == '.') i++;

				snprintf(tmp.ut_host, UT_HOSTSIZE,"%03d:%c.%s", port,
					ai->proto, p);
			} else
				snprintf(tmp.ut_host, UT_HOSTSIZE, "%03d:%c.%s", 
					port, ai->proto, ai->message);			
				//strncpy(tmp.ut_host, "%03d:", port);
#else
//new format
			if (ai->address > 0)
				strcpy(ip, dotted(ai->address));
			else
				strcpy(ip, "");
			
			if ((ai->address) && (islogin > P_INCOM))
				m = ai->conn_info;
			else
				m = ai->message[0];
						
			snprintf(tmp.ut_host, UT_HOSTSIZE, "%03d:%c:%s:%s",
				port, ai->proto, ip, m);
#endif

#ifdef __linux__
			tmp.ut_addr = ai->address;
#endif
			setutent();
			pututline(&tmp);
			endutent();
			if (lineconf[port].syswtmp &&
			    (fp = fopen(WTMP_FILE, "a")) != NULL) {
				fwrite(&tmp, sizeof(tmp), 1, fp);
				fclose(fp);
			}
		}
	}

#if 0 /* To be implemented */
	if (lineconf[port].radutmp == 0)
		return 0;

	/*
	 *	Now update the /var/log/radutmp file.
	 */
	memset(&rtmp, 0, sizeof(rtmp));
	strncpy(rtmp.login, username, sizeof(rtmp.login));
	rtmp.nas_port = 
	memcpy(rtmp.session_id, ai->session, 16);
	rtmp.nas_address = ai->localip;
	rtmp.framed_address = ai->addres;
	rtmp.proto = ai->proto;
	rtmp.time = time(NULL);
	rtmp.type = islogin ? P_LOGIN : P_IDLE;
#endif

	return 0;
}

int update_status(int port, char *login, char *message)
{

struct auth status_ai;

	status_ai.proto = 'I';
	status_ai.address = 0;
	strcpy(status_ai.login, login);
	strcpy(status_ai.message[0], message);
	
	strncpy(status_ai.login, login, sizeof(status_ai.login) - 1);
	status_ai.login [sizeof(status_ai.login) - 1] = '\0';
	strncpy(status_ai.message[0], message, sizeof(status_ai.message[0]) - 1);
	status_ai.message[0] [sizeof(status_ai.message[0]) - 1] = '\0';

	return ( update_utmp(port, &status_ai, P_INCOM) );
	
}
